{--------------------------------------------------------------------------------------------------------------------------

Conway's Game of Life implmented in Haskell - Peter Henderson - s1117205 - peter.henderson@gmx.com



I decided to pick something a little challenging for this competition to stretch my skills somewhat, so instead of creating
a turtle graphic I jumped in at the deep end with OpenGL and GLUT. *gasp* D:

Turned out to be a great learning excercise, good fun too, and not as bad as I was expecting :)

*****
I advise against using a board size above around 400.. processing time increases to an unacceptable level after that, making the 
program unresposive. Do so at your own peril!
*****

Dependancies: OpenGL, GLUT

---------------------------------------------------------------------------------------------------------------------------}




import GOLTypes    
import System.Random (randoms, mkStdGen)
import Data.String -- only for 'read'
import Data.IORef 
import System.Exit (exitSuccess)

import Graphics.Rendering.OpenGL
import Graphics.UI.GLUT


{------------------------------------------------------------------------------------------------------------------
Generates a square, n x n board (where n>=5) with random entries of 'Alive' or 'Dead' based on the seed given.
Same 'seed' and 'size' generates the same board, so favourite patterns can be found and repeated!

Puts rows and columns of all 'Dead' cells around the outside of the board, that will never be updated,
in order to give an edge to the board when carrying out generation calculations and displaying the result.
-------------------------------------------------------------------------------------------------------------------}

makeBoard :: Int -> Int -> Board
makeBoard size seed | size >= 5     = drow ++ [[Dead]++[f x | x <- randomList y]++[Dead] | y <- [1..size]] ++ drow
                    | otherwise  = error "Cannot make board of size less than 5"

               where f n | n `mod` 2 == 0 = Dead
                         | otherwise      = Alive
                     drow         = [(replicate (size+2) Dead)]
                     randomList y = (take size (rList (seed+y)))
                 

rList :: Int -> [Integer]                    
rList n = randoms (mkStdGen n)






{----------------------------------------------------------------------------------------------------------------------------------------------------
The main function of the whole program, this recursively goes through each cell of the board (exculding the first and last row, and the first and 
last cell in each row), updates its state accoring to the rules listed below, and outputs the new board one "generation" advanced.

Rules (taken from Wikipedia):
      Any live cell with fewer than two live neighbours dies, as if caused by under-population.
      Any live cell with two or three live neighbours lives on to the next generation.
      Any live cell with more than three live neighbours dies, as if by overcrowding.
      Any dead cell with exactly three live neighbours becomes a live cell, as if by reproduction.

These rules can be simplified to two checks to save on computation:
      If the cell has exactly 3 live neighbours, it will become 'Alive'.
      If the cell has exactly 2 live neighbours, it remains the same as the previous generation.
      Otherwise, the cell becomes 'Dead'.
------------------------------------------------------------------------------------------------------------------------------------------------------}

advGeneration :: Int -> Board -> Board
advGeneration size board = drow ++ [(advRow prev row next size []) | (prev,row,next) <- zip3 board (drop 1 board) (drop 2 board)] ++ drow
              where drow = [(replicate (size+2) Dead)]



-- advRow was originally implemented using append without reverse, ie. building up the output Row by appending an item to the end of the list.
-- Obviously, this is much more costly than using cons and reversing the output list. This way, the number of steps required to add an item to the output is always ~1.
-- This implementation is a much faster version for larger board sizes (eg. for a board size of 400 this implementation is just under 3x faster than the original.

-- There are probably further improvements to be made on this though..

advRow ::  Row -> Row -> Row -> Int -> Row -> Row
advRow _ _ _ 0 output     = Dead:(reverse(Dead:output))  -- To preserve the column of 'Dead' at the right and left hand sides of the board.
advRow (a1:a2:a3:xs) (b1:y:b3:ys) (c1:c2:c3:zs) steps output | liveNeighbours == 3 = advRow (a2:a3:xs) (y:b3:ys) (c2:c3:zs) (steps-1) (Alive:output)
                                                             | liveNeighbours == 2 = advRow (a2:a3:xs) (y:b3:ys) (c2:c3:zs) (steps-1) (y:output)
                                                             | otherwise           = advRow (a2:a3:xs) (y:b3:ys) (c2:c3:zs) (steps-1) (Dead:output)

                                           where liveNeighbours = foldr (+) 0 (map live nList)
                                                 nList = a1:a2:a3:b1:b3:c1:c2:c3:[]   
  
  
live :: Cell -> Int
live Alive = 1
live Dead  = 0


                              





{---------------------------------------------------------------------------------------------------------

*Rendering Functions*

Uses OpenGL and GLUT




Only renders squares in locations that have an 'Alive' cell in a hope to save on computation time. 
The main bottleneck, however, is the advRow function, so this probably makes little to no difference...

---------------------------------------------------------------------------------------------------------}

main :: IO()
main = do putStr "An implementation of \"Conway's Game of Life\" in Haskell. \n\nWritten by Peter Henderson for the 2011 Inf1 FP Programming Competition. \nPress ENTER to exit the program when you are finished watching the output. \n\nWhat size of board (Minimum of 5) : "

          putStr "\n" --anomoly required for the compiled version not to freak out.. seems it doesnt like outputing the line previous to a getInput call... strange

          (size,seed) <- getInput
          let board = makeBoard size seed
          boardm <- newIORef board

          (progname,_) <- getArgsAndInitialize
          initialWindowSize $= Size 600 600
          initialWindowPosition $= Position 50 50
          initialDisplayMode $= [DoubleBuffered]
          actionOnWindowClose $= Exit

          createWindow "Conway's Game of Life in Haskell - 2011 Inf1FP Competition Entry - Peter Henderson"

          displayCallback $= display boardm size
          keyboardMouseCallback $= Just keyboardMouse

          mainLoop



getInput :: IO(Int,Int)
getInput = do a <- getLine
              let size = read a
              putStr "What seed value? : "

              putStr "\n" --same odd error as before...

              b <- getLine
              let seed = read b
              return (size,seed)

----------------------------------

display :: (HasSetter s, HasGetter s) => s Board -> Int -> IO ()
display boardm size = do
        clear [ColorBuffer]
        board <- get boardm
        color $ Color3 (0.5::GLfloat) (1.0) 0 --I picked green, you can change the colour here if you desire. (r,g,b)
        renderPrimitive Quads $ mapM_ (\(x,y,_) -> vertex $ Vertex3 x y 0.0) (cellsAlive board size)
        boardm $=! advGeneration size board
        swapBuffers
        postRedisplay Nothing



cellsAlive :: Board -> Int -> [(GLfloat,GLfloat,GLfloat)]
cellsAlive board n = concat [points cs [(x,y,0) | (cell,y) <- zip row span, isAlive cell] | (row,x) <- zip board span]
    where cs = (1/(size/2))
          size = fromIntegral (n+2)
          span = [(-(size/2))..(size/2)]

points :: Fractional t => t -> [(t, t, t1)] -> [(t, t, t1)]
points cs ((x,y,z):xs) = (x*cs,y*cs,z) : (x*cs,(y+1.0)*cs,z) : ((x+1.0)*cs,(y+1.0)*cs,z) : ((x+1.0)*cs,y*cs,z) : points cs xs
points _ []            = []



isAlive :: Cell -> Bool
isAlive Alive = True
isAlive Dead  = False


---------------

keyboardMouse :: Key -> KeyState -> t -> t1 -> IO ()
keyboardMouse (Char '\r') Down _ _ = do exitSuccess
keyboardMouse _ _ _ _ = do return ()
